/******************************************************************************
 * Copyright (C) 2003 David Tam, Catalin Drula, Winnie Tsang
 * tamda@eecg.toronto.edu
 * catalin@cs.toronto.edu
 * wtsang@dgp.toronto.edu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *****************************************************************************/
#include <linux/dvs.h>
#include <linux/module.h>

/*
 * Here is a small static array for holding the quantum usage history.
 * This will be used as a circular buffer.
 */
dvs_quantum_usage_t dvs_qu_array[DVSBUCKETS];

/* The "current pointer" of the above circular buffer */
int dvs_qu_index = 0;

/* The current cpu speed */
/* initialize to 100 (I assume the cpu starts at full speed */
int dvs_speed = 100;

EXPORT_SYMBOL_GPL(dvs_qu_array);
EXPORT_SYMBOL_GPL(dvs_qu_index);
EXPORT_SYMBOL_GPL(dvs_speed);




/******************************************************************************
 ******************************************************************************
 * This BIG BIG section is included only for the embedded version of DVS.
 *
 */
#ifdef DVS_EMBEDDED

#if 0
/* Global variable for determining debug print out occurence */
int dbg_ctr = 0;
#endif /* 0 */

/* Variable to match loadable kernel module version of DVS */
int interval_length_jiffies = DVSFREQ;

/* When == 0, then the DVS algorithm is run */
int dvs_wait = DVSFREQ;


/*****************************************************************************/
/* This subsection contains functions to process kernel dvs data             */
/*****************************************************************************/

/* Function that returns the avg given time slice size for task p */
int avg_orig_time_slice(task_t *p) {
        int i;
	int ctr = 0; // Counts number of nonblank entries
	int sum = 0;
	int avg;

	for (i = 0 ; i < TASKLOADSIZE; i++) {
		/* Skip blank entries */
		if (p->task_load_array[i].orig_time_slice == 0) {
			continue;
		}

		ctr++;
		sum += p->task_load_array[i].orig_time_slice;
	}		

	/* Guard against division by zero */
	if (ctr != 0) {
		avg = sum / ctr;
		if (avg % ctr != 0) { avg++; } /* Round up if necessary */
	} else {
		avg = p->orig_time_slice;
	}

	return avg;
}


/***********************************************************
 * FIXME: Real ugly.
 * The following declarations are ripped from kernel/sched.c
 * I need them for the get_idlecycles_future() function.
 * I don't know of a better solution.
 *
 *
 */

/*
 * These are the runqueue data structures:
 */

#define BITMAP_SIZE ((((MAX_PRIO+1+7)/8)+sizeof(long)-1)/sizeof(long))

typedef struct runqueue runqueue_t;

struct prio_array {
	int nr_active;
	unsigned long bitmap[BITMAP_SIZE];
	struct list_head queue[MAX_PRIO];
};

/*
 * This is the main, per-CPU runqueue data structure.
 *
 * Locking rule: those places that want to lock multiple runqueues
 * (such as the load balancing or the thread migration code), lock
 * acquire operations must be ordered by ascending &runqueue.
 */
struct runqueue {
	spinlock_t lock;
	unsigned long nr_running, nr_switches, expired_timestamp,
			nr_uninterruptible;
	task_t *curr, *idle;
	struct mm_struct *prev_mm;
	prio_array_t *active, *expired, arrays[2];
	int prev_cpu_load[NR_CPUS];
#ifdef CONFIG_NUMA
	atomic_t *node_nr_running;
	int prev_node_load[MAX_NUMNODES];
#endif
	task_t *migration_thread;
	struct list_head migration_queue;

	atomic_t nr_iowait;
};



extern struct runqueue per_cpu__runqueues;

#include <linux/percpu.h>
#define this_rq()		(&__get_cpu_var(runqueues))

/*
 *
 *
 * FIXME: Real ugly.
 * The above declarations are ripped from kernel/sched.c
 *******************************************************/

/* Forware function declaration */
int task_load_calc(task_t *p);

/* Provide information about upcoming jobs in the run queue */
/* Returns a value in jiffies. */
int get_idlecycles_future(void) {
	int idx = 0;
	runqueue_t *rq = this_rq();
	prio_array_t *array = rq->active;
	int flag_checked_active = 0;
	unsigned int total_given = 0;
	unsigned int total_idle = 0;

	/**********************************************************************
	 * Account for remainder of current job
	 *
	 */

	float est_load;
	unsigned int est_runtime;
	unsigned int consumed_sofar;

	/* Adjust utilization for current cpu speed
	 * Pseudo-code:
	 * est_load = (task_load_calc(current) / 100.0) / (dvs_speed / 100.0);
	 * Real code below: */
	est_load = (float) task_load_calc(current) / dvs_speed;

	/* Calculate estimated run time for current job at current speed (since
	 * it is currently running at the current speed). */
	est_runtime = est_load * current->orig_time_slice;

	consumed_sofar = current->orig_time_slice - current->time_slice;

	/* Apply only if task is not taking longer than expected */
	if (est_runtime >= consumed_sofar) {
		unsigned int est_tobe_consumed = est_runtime - consumed_sofar;
		/* Normalize future consumption back to a full speed cpu */
		est_tobe_consumed = est_tobe_consumed * (dvs_speed / 100.0);

		/* Only if there is enough time left do we add the remaining to
		 * the idle time variable */
		if (current->time_slice > est_tobe_consumed) {
			total_idle = current->time_slice - est_tobe_consumed;
		}
	}

	total_given = current->time_slice;

	/*
	 *
 	 * Account for remainder of current job
	 *********************************************************************/

	/* Traverse run queue and find a candidate priority list */
	while (1) {
		struct list_head *headlist, *curlist;
		task_t *curtask;


		/* Find highest priority list with a job */
		if (idx == 0) {
			idx = sched_find_first_bit(array->bitmap);
		} else {
			idx = find_next_bit(array->bitmap, MAX_PRIO, idx);
		}

		/* If run queue has no more jobs for our calculation */
		if (idx >= MAX_PRIO) {
			/* Detect if already tried both active and inactive
			 * queues */
			if (flag_checked_active == 1) {
				/* Account for lack of jobs.
				 * Scale idle to interval size.
				 * Pseudo-code:
				 * total_idle = total_idle / (total_given /
				 *                  interval_length_jiffies);
				 * Real code: */
				total_idle = total_idle *
					interval_length_jiffies / total_given;

				/* The following proved to be too conservative
				 * and gave way too much idle time in some 
				 * initial tests. 
				 * Add remaining idle time to total_idle
				 * total_idle += interval_length_jiffies -
				 *               total_given;
				 */
				break;
			}

			/* Have not tried inactive queue yet.
			 * Set flag and try inactive queue */
			flag_checked_active = 1;
			idx = 0;
			array = rq->expired;
			continue;
		}

		/* Go to target priority list */
		headlist = array->queue + idx;
		curlist = headlist->next;
		curtask = list_entry(curlist, task_t, run_list);

		/* Traverse target priority linked-list */
		while (1) {
			int idlepercent, cur_idle;
			int avg_tslice = avg_orig_time_slice(curtask);

			total_given += avg_tslice;

			/* Prevent overflow. Leads to a conservative idle
			 * estimate */
			if (total_given > interval_length_jiffies) { break; }

			/* Accumulate idle jiffies */
			idlepercent = 100 - task_load_calc(curtask);
			cur_idle = (idlepercent / 100.0) * avg_tslice;
			total_idle += cur_idle;

			/* Next job on this priority list.
			 * For some reason, the following line was incorrect:
			 * cur = list_entry(cur->run_list.prev, task_t,
			 *                  run_list);
			 * The following line and a few appropriate mods solved
			 * the problem. */
			curlist = curlist->next;
			if (curlist == headlist) { break; }
		}

		if (total_given >= interval_length_jiffies) { break; }

		/* Continue searching starting at next priority */
		idx++;

		/* If run queue has no more jobs for our calculation */
		if (idx >= MAX_PRIO) {
			/* Detect if already tried both active and inactive
			 * queues */
			if (flag_checked_active == 1) {
				/* Account for lack of jobs.
				 * Scale idle to interval size.
				 * Pseudo-code:
				 * total_idle = total_idle / (total_given /
				 *                  interval_length_jiffies);
				 * Real code: */
				total_idle = total_idle *
					interval_length_jiffies / total_given;

				/* The following proved to be too conservative
				 * and gave way too much idle time in some
				 * initial tests. 
				 * Add remaining idle time to total_idle
				 * total_idle += interval_length_jiffies -
				 *               total_given;
				 */
				break;
			}

			/* Have not tried inactive queue yet.
			 * Set flag and try inactive queue */
			flag_checked_active = 1;
			idx = 0;
			array = rq->expired;
		}

	}

	if (total_idle > interval_length_jiffies) { 
		total_idle = interval_length_jiffies;
	}
	return total_idle;
}

/* Right now interval is in the range of [-1,1].
 * -1 specifies that you want to use the "pseudo-future" algorithm.
 * For now, I am assuming you mean idle jiffies rather than real cycles.
 */
int get_idlecycles(int interval) {
	int qu_index = dvs_qu_index; /* Index into dvs_qu_array[] */
	int total_jiffies = 0;
	unsigned long long total_idlecycles = 0;
	int divisor;
	int rc;
	int jobctr = 0;
	int skipctr = 0;


	/* Requested information for a "pseudo-future"-based algorithm */
	if (interval == -1) {
		return get_idlecycles_future();
	}


	/* ************************************************** */
	/* Requested information for a "past"-based algorithm */


	/* Work backwards from current index into dvs_qu_array[] */
	while (1) {
		jobctr++;
		qu_index--;
		/* Handle index wrap-around */
		if (unlikely(qu_index < 0)) { qu_index = DVSBUCKETS - 1; }

		/* Prevent potential infinite loop due to wrap-around on
		 * dvs_qu_array[] */
		if (qu_index == dvs_qu_index) { break; }

		/* If not in target interval to gather idle cycles */
		if (total_jiffies < interval_length_jiffies * interval) {
			skipctr++;
			total_jiffies += 
				dvs_qu_array[qu_index].orig_time_slice;

			/* If crossed into target interval */
			if (unlikely(total_jiffies > interval_length_jiffies *
				     interval)) {
				/* Taking an appropriate portion of idle cycles
				 * from nonspillage. */
				int spillage = total_jiffies -
					(interval_length_jiffies * interval);
				int nonspillage =
					dvs_qu_array[qu_index].orig_time_slice
					- spillage;
				/* Do the follow calc, which involves a 64-bit
				 * division.
				 * total_idlecycles += (nonspillage *
				 * dvs_qu_array[qu_index].ns_left) /
				 * dvs_qu_array[qu_index].orig_time_slice;
				 */
				unsigned long long number = nonspillage *
					dvs_qu_array[qu_index].ns_left;
				do_div(number,
				      dvs_qu_array[qu_index].orig_time_slice);
				total_idlecycles += number;

			}

			continue;
		}
		
		/*
		 * From this part onwards, we are in the target interval to
		 * gather idle cycles
		 */

		/* Handle when near interval boundary */
		if (unlikely(total_jiffies +
			     dvs_qu_array[qu_index].orig_time_slice >
			     interval_length_jiffies * (interval + 1))) {
			/* Taking an appropriate portion of idle cycles
			 * from boundary element */
			int remainder = (interval_length_jiffies *
					 (interval + 1)) - total_jiffies;
			if (unlikely(remainder <= 0)) { break; }
			/* Do the follow calc, which involves a 64-bit
			 * division.
			 * total_idlecycles += (remainder *
			 * dvs_qu_array[qu_index].ns_left) /
			 * dvs_qu_array[qu_index].orig_time_slice;
			 */
			unsigned long long number = remainder *
				dvs_qu_array[qu_index].ns_left;
			do_div(number, dvs_qu_array[qu_index].orig_time_slice);
			total_idlecycles += number; 
			break;
		} else {
			/* Handle when not near interval boundary */
			total_jiffies +=
				dvs_qu_array[qu_index].orig_time_slice;
			total_idlecycles += dvs_qu_array[qu_index].ns_left;

			/* If at interval boundary */
			if (unlikely(total_jiffies >=
				     interval_length_jiffies *
				     (interval + 1))) {
				break;
			}
		}
	}

	/* Replaced the following with 3 code lines below to allow for 64-bit
	 * division.
	 * rc = NS_TO_JIFFIES(total_idlecycles);
	 */
	divisor = 1000000000 / HZ;
	do_div(total_idlecycles, divisor);
	rc = total_idlecycles;

	if (unlikely(rc > interval_length_jiffies)) {
		rc = interval_length_jiffies;
	}
	return rc;
}

/* Right now interval is in the range of [-1,1].
 * -1 specifies that you want to use the "pseudo-future" algorithm.
 * For now, I am assuming you mean jiffies rather than real cycles.
 */
int get_runcycles(int interval) {
	return interval_length_jiffies - get_idlecycles(interval);
}

/* Estimated excess cycles based on load estimator.
 * Right now interval is in the range of [-1,1].
 * -1 specifies that you want to use the "pseudo-future" algorithm.
 * For now, I am assuming you mean jiffies rather than real cycles.
 */
int get_excesscycles(int interval) {
	/* I don't know. Return 0 for now. */
	return 0;
}

/* Function that sums up the N load records from the
 * task_load_array and returns the average (an integer between 0 and 100).
 * NOTE: The average is for a full speed cpu.
 * N == TASKLOADSIZE */
int task_load_calc(task_t *p) {
        int i=0;
	int ctr = 0; // Counts number of nonblank entries
        int avg_load = 0;  // compute avg_load. Integer between 0 and 100.
        
	/* Compute the load for ith bin */
        for (i=0; i<TASKLOADSIZE; i++){
		int divisor = 1000000000 / HZ;
		unsigned long long int tmp;
		unsigned int jiffies_left;
		unsigned int run_jiffies;
		float util;
		int util_int;

		/* Skip blank entries */
		if (p->task_load_array[i].orig_time_slice == 0) {
			continue;
		}

		ctr++;

		/* convert ns_left to jiffies */
		tmp = p->task_load_array[i].ns_left;
		do_div(tmp,divisor);
		jiffies_left = (unsigned int) tmp;

		/* Skip if consumed only "bonus" jiffies.
		 * Sometimes, tasks are given a priority boost ("bonus"
		 * jiffies) for sleeping alot. */
		if (jiffies_left >= p->task_load_array[i].orig_time_slice) {
			continue;
		}

		/* Calculate run time jiffies */
		run_jiffies = p->task_load_array[i].orig_time_slice -
			jiffies_left;

		/* Calculate utilization. Get a real between 0 and 1.
		 * I need the float cast because result will be 0.xxxxxxx,
		 * which would otherwise get truncated to 0 before being
		 * assigned to util.
		 */
		util = (float) run_jiffies /
			p->task_load_array[i].orig_time_slice;

		/*****************************************************
		 * Pseudo-code for next few steps to normalize utilization to a
		 * full speed cpu.
		 *
		 * // Convert to an integer between 0 and 100
		 * util_int = util * 100;
		 *
		 * // Convert integer speed to a real between 0 and 1
		 * speed = p->task_load_array[i].speed / 100;
		 *
		 * // Normalize value to a full speed cpu
		 * // Result is an integer between 0 and 100
		 * util_int = util_int * speed;
		 * 
		 * Optimized version in real-code below.
		 * Normalize utilization to a full speed cpu.
		 *****************************************************/
		util_int = util * p->task_load_array[i].speed;

		/* accumulate value */
		avg_load+= util_int;
        }

	/* take the sum of the load and compute average */
	/* Guard against division by zero */
	if (ctr != 0) {
		avg_load = avg_load/ctr;
		/* Round up if necessary */
		if (avg_load % ctr != 0) {
			avg_load++;
		}
	} else {
		avg_load = 0;
	}

        return avg_load;
}


/*****************************************************************************/
/* This subsection contains the DVS core algorithms                          */
/*****************************************************************************/

#define CONSTANT 1    /* CONSTANT for Peak Algorithm */
#define MIN_SPEED 10	 /* minimum percentage of speed */
#define HYST_MIN  50  /* For stability in past/future algorithm. orig=50*/
#define HYST_MAX  70  /* For stability in past/future algorithm. orig=70*/
#define PAST_INCR 20   /* Size of increment in past/future algorithm */
#define PAST_BASESUB HYST_MIN + 10 /* Size of base number used in decrement
				      operation in past/future alg. orig=60 */

/* Variables to more closely match loadable kernel module version of DVS */
int hyst_min = HYST_MIN;
int hyst_max = HYST_MAX;
int past_incr = PAST_INCR;
int past_basesub = PAST_BASESUB;

/* structure for store interval statistics */
struct dvs_struct{
	int run_percent;
	int speed;
};

/* stores the current */
struct dvs_struct dvs_current = { .run_percent = 0, .speed = 100 };

/* Main function implementing the peak algorithm */
/* PEAK */
/* Compute the speed and predicted run_percent and stores these values
   into *dvs_current struct, return speed (int) */
int peak(void) {

	/* Declare local variables */
	int run_cycle;
	int idle_cycle;
	int excess_cycle;
	int excess_percent;
	int next_excess_cycle;
	int run_percent[2];

	/* Get Pt-1 run percent */
	run_cycle=get_runcycles(1);
	idle_cycle=get_idlecycles(1);
	excess_cycle=get_excesscycles(1);
#if 0
	if (dbg_ctr >= 20) {
		printk(KERN_ERR "run_cyc1 = %d, idle_cyc1 = %d, excess_cyc1 = \
%d\n",
		       run_cycle, idle_cycle, excess_cycle);
	}
#endif /* 0 */
	run_percent[1] = (run_cycle + excess_cycle)*100/(run_cycle +
							 idle_cycle);
	
	/* Get Pt run percent */
	run_cycle=get_runcycles(0);
	idle_cycle=get_idlecycles(0);
	excess_cycle=get_excesscycles(0);
#if 0 
	if (dbg_ctr >= 20) {
		printk(KERN_ERR "run_cyc0 = %d, idle_cyc0 = %d, excess_cyc0 = \
%d\n",
		       run_cycle, idle_cycle, excess_cycle);
	}
#endif /* 0 */
	run_percent[0] = (run_cycle + excess_cycle)*100/(run_cycle + 
							 idle_cycle);

	/* Compute excess cycle outgoing */
	next_excess_cycle = ((excess_cycle+run_cycle)*100-dvs_current.speed*
			     (run_cycle+idle_cycle))/100;
	if (next_excess_cycle < 0) {
		next_excess_cycle = 0;
	}
	excess_percent = next_excess_cycle*100/(run_cycle + idle_cycle);

	/* Prediction, set dvs_current's run percent based on past two
	 *  intervals statistics*/
	/* Pt > Pt-1 */
	if (run_percent[0] > run_percent[1]) {
		if (run_percent[1] > 10) {
			dvs_current.run_percent = run_percent[1];
		}
		else {
			dvs_current.run_percent = 10;
		}
	}
	/* Pt < Pt-1 */
	else if (run_percent[0] < run_percent[1]) {
		if (run_percent[1] > 10) {
			dvs_current.run_percent = run_percent[0];
		}
		else {
			dvs_current.run_percent = 10;
		}
	}
	/* Pt = Pt-1 */
	else {
		if (run_percent[0]==100) {
			dvs_current.run_percent = 40;
		}
		else{
			dvs_current.run_percent=run_percent[0];
		}
	}
	
	/* Set Speed */
	dvs_current.speed = dvs_current.run_percent + CONSTANT*excess_percent;
	if (dvs_current.speed > 100) {
		dvs_current.speed = 100;
	}
	if (dvs_current.speed < MIN_SPEED) {
		dvs_current.speed = MIN_SPEED;
	}
	return dvs_current.speed;
}

/* Main function implementing the past and future algorithms */
/* PAST/FUTURE */
/* Compute the speed and predicted run_percent and stores these values
 * into *dvs_current struct and return point to decision too.
 * Parameters:
 * interval == -1 use pseudo-future algorithm
 * interval == 0  use past algorithm
 */
int futurepast(int interval) {
	/* Declare local variables */
	int run_cycle;
	int idle_cycle;
	int excess_cycle;
	int excess_percent;
	int next_excess_cycle;

	/* Check parameter boundary */
	if (interval < -1 || interval > 0) {
		return -1;
	}

	/* Get Pt run percent */
	run_cycle=get_runcycles(interval);
	idle_cycle=get_idlecycles(interval);
	excess_cycle=get_excesscycles(interval);

	dvs_current.run_percent = (run_cycle + excess_cycle)*100/
				   (run_cycle + idle_cycle);

#if 0
	if (dbg_ctr >= 20) {
		printk(KERN_ERR "futurepast(): run%%=%d run=%d idle=%d excess=\
%d\n",
		       dvs_current.run_percent, run_cycle, idle_cycle,
		       excess_cycle);
	}
#endif /* 0 */
	
	/* Compute excess cycles */
	next_excess_cycle = ((excess_cycle+run_cycle)*100-
			     dvs_current.speed*(run_cycle+idle_cycle))/100;
	if (next_excess_cycle < 0) {
		next_excess_cycle = 0;
	}
	excess_percent = next_excess_cycle*100/(run_cycle + idle_cycle);

	/* Set Speed */
	if (excess_cycle > idle_cycle) {
		dvs_current.speed = 100;
	}
	else if (dvs_current.run_percent > hyst_max) {
		dvs_current.speed = dvs_current.speed + past_incr;
	}
	else if (dvs_current.run_percent < hyst_min) {
		dvs_current.speed = dvs_current.speed -
			(past_basesub - dvs_current.run_percent);
	}
	if (dvs_current.speed > 100) {
		dvs_current.speed = 100;
	}
	if (dvs_current.speed < MIN_SPEED) {
		dvs_current.speed = MIN_SPEED;
	}
	return dvs_current.speed;
}


/******************************************************************************
 * This subsection contains functions called by sched.c
 *****************************************************************************/

/* Runs the appropriate CPU governor
 * This means that the DVS loadable kernel module is not being used. */
void dvs_runalg(task_t *p) {
	static int old_dvs_speed = 100;
	/* Used in embedded version since there is no freq/voltage scaling
	 * capability */
	int fake_dvs_speed;

	/* Run either futurepast() or peak() */
	fake_dvs_speed = futurepast(0);

	/* Print only when speed changes */
	if (old_dvs_speed != fake_dvs_speed) {
		printk(KERN_ERR "dvs_runalg(): run%%=%d fake_dvs_speed:=%d\n",
		       dvs_current.run_percent, fake_dvs_speed);

	}
	old_dvs_speed = fake_dvs_speed;

#if 0
	/* Print once every 20 times */
	if (dbg_ctr >= 20) {
		printk(KERN_ERR "util = %d, fake_dvs_speed = %d\n",
		       dvs_current.run_percent, fake_dvs_speed);
		dbg_ctr = 0;
	}
	dbg_ctr++;
#endif /* 0 */

}


#endif /* DVS_EMBEDDED */
/*
 *
 * This BIG BIG section is included only for the embedded version of DVS.
 *****************************************************************************
 *****************************************************************************/
